=begin
PipeAlongPath
TIG (c) 2005 - 2013
Description : Creates circular faced 'followme' pipe extrusions along a path - 
Use it to make pipes, ducts etc, starting from a selection of joined edges
Usage : 
Select joined lines, arcs, circles, curves, etc.
Select "Pipe Along Path" from the Plugins menu. 
In the dialog choose:-
Units (mm/inch):
Outside diameter: 
Inside diameter: 
Number of segments: 
Cpoints?: [Yes/No, default=true to add cppoints at the pipe's vertices]
Cline layer: [layer for the path, default="XCLINE", make blank for 'none']:  
In group?: [Yes/No to move selected path into pipe's group, default=true]
OK.  
Setting are remembered during a session.
The extrusion is made*.
The extrusion is grouped so it does not interact with adjacent surfaces - 
afterwards just explode it if appropriate. 
Edit it to intesect with model and tidy up to make tees etc...
The pipe also has construction points added at vertices if set Cpoints=true.  
These can be used for snapping, if not wanted 'erase construction geometry' 
will remove them globally or just within a group that you are editing.
One Undo to remove the construction points, a second to Undo the Pipe itself.
The default for extrusion face segments is 24, the minimum is 3.
The alignment is always 'Centroid' - along the pipe's centre line.
The diameter is always measured square to the vector of the 
first path's line.  A diameter that is less than the length of pieces of path 
edge might give correct but unexpected extrusions, which might need manually 
tidying...
Closed loop paths are fully extruded in a loop.
Multiple arcs and other complex 3D paths might give unexpected results.  
Note that SketchUp can't handle very small faces in its FollowMe mode - so any 
Arc bends of 8" radius or less that have the same radius for the pipe (o/d) 
applied will almost certainly cause a crash / "bug splat" and are trapped out 
BUT note that similar radii in 'welded' Curves are NOT easily trappable and so 
they may always cause a crash with a "bug-splat", so avoid using these type of 
path with small radii bends/pipes - or you can keep it but use 'scale' as
explained below...
IF you must have this matching small diameter pipe and arc or welded elbow 
bends then to get it to work you can make the path a 
group, scale it by a factor so it's larger (say x10), then 
edit it and use this tool within it applying the diameters x10, 
after it's all made scale the group back down by x0.1 and 
explode it and it'll all be OK.
#
Both diameters cannot be zero and will return an error.
If one of the diamaters is zero then you get a 'tube' rather 
than a pipe.
If the diamaters entered are equal the inside one is taken as 
zero and you will then get a 'tube' rather than a pipe.
If the inside's diameter is greater than the outside's then 
they are swapped.
A branching path returns an error as 'FollowMe' can't decide 
which path to take.
Selected edges that are not joined (i.e. they don't all have 
common ends-starts) will not be extruded - only the first edge 
or joined group of edges will extrude.

Version :
1.0  18/9/5	first release.
1.1  18/9/5	visual segmenting of arced sections etc addressed.
1.2  9/2/6  centrepoints added at vertices of non-looped paths,
			occasional reversed faces on single 'up' line fixed.
1.3  16/2/6 undo of cpoints fixed, diam <8" with matching arc 
			elbow bend radius bug splat trapped (but not curves).
1.4  2/5/6	group name -> OD=xx ID=nn with " or mm as units.
1.5  3/5/6	Transposed ID/OD fixed (sorry!).
1.6 12/5/7  @error=0 ### fix
1.7 20121019 Rehashed to modern standards, dialog based options to add cpoints 
			 at nodes, put selected path onto a specified layer and move the 
			 path inside the pipe-group.
1.8 20130115 Fixed typo glitch when path was lone vertical downward line.
=end
###
require 'sketchup.rb'
###

module TIG

###

def self.pipe_along_path()

model = Sketchup.active_model
entities = model.active_entities
ss = model.selection
layers=model.layers
###
### show VCB and status info
Sketchup::set_status_text("Pipe Along Path... PARAMETERS...", SB_PROMPT)
Sketchup::set_status_text(" ", SB_VCB_LABEL)
Sketchup::set_status_text(" ", SB_VCB_VALUE)
###
if ss.empty?
  Sketchup::set_status_text("Pipe Along Path... NO SELECTION ! ", SB_PROMPT)
  UI.messagebox("No Selection to Extrude Pipe.")
  return nil
end
### trap for small radius arcs
aflag=false
arads=[]
for s in ss
  if s.is_a?(Sketchup::Edge)
    if s.curve
      if s.curve.is_a?(Sketchup::ArcCurve)
        if s.curve.radius < 200.mm
          aflag=true
          arads=arads.push(s.curve.radius)
        end
      else
        for e in s.curve.edges
          if e.curve.is_a?(Sketchup::ArcCurve)
            if e.curve.radius < 200.mm
              aflag=true
              arads=arads.push(e.curve.radius)
            end
          end
        end#for
      end
    end
  end
end#for
### ####################################################################
### dialog #############################################################
units = ["mm","inch"]
enums = [units.join("|"),"","","","Yes|No","","Yes|No"]
if not @diamOutIn

### Changed JMcC 2013-06-12 to default to scaffold tube size (1.9in or 48.26mm), and imperial units
###   values = ["mm", 110.0, 100.0, 24, "Yes","XCLINE","Yes"]
   values = ["inch", 1.9, 0.0, 8, "Yes","CLINE","No"]
else
   values = [@unitsIn, @diamOutIn, @diamInIn, @psegmentsIn, @cpoints, @clineLayer, @inGroup]
end
prompts = ["Units (mm / inch) : ", "Outside Diameter: ", "Inside Diameter: ", "Number of Circle Segments: ", "Cpoints at Nodes? ", "Cline Layer: ", "Path into Pipe Group? "]
results = inputbox(prompts, values, enums, "Pipe Parameters")
return nil unless results ### i.e. the user cancelled the operation
@unitsIn, @diamOutIn, @diamInIn, @psegmentsIn, @cpoints, @clineLayer, @inGroup = results
if @diamOutIn == 0 and @diamInIn == 0 ### can't BOTH be 0 ###
   @diamOutIn = 110.0
   UI.messagebox("Two Zero Diameters NOT allowed ! ")
   return nil
end
if @diamOutIn == @diamInIn ### can't be same so Inner=0 ###
   @diamInIn = 0
   UI.messagebox("Equal Diameters NOT allowed !\nInner Diameter is reset to Zero. ")
end
if @diamInIn > @diamOutIn ### Inner can't be bigger ###
   diamInIntemp = @diamInIn
   @diamInIn = @diamOutIn
   @diamOutIn = diamInIntemp
   UI.messagebox("Inner Diameter NOT allowed to be bigger than Outer!\nInner is reset to be the smaller value. ")
end
if @psegmentsIn < 3
   @psegmentsIn = 24
   UI.messagebox("Fewer than 3 Faces NOT allowed !\nDefaulting to 24. ")
end
segs = @psegmentsIn
radiusO = @diamOutIn / 2
radiusI = @diamInIn / 2
### - do units stuff - ###
if @unitsIn == "mm"
   radiusI = radiusI.mm
   radiusO = radiusO.mm
   txtO=@diamOutIn.to_s+"mm"
   txtI=@diamInIn.to_s+"mm"
else
   radiusI = radiusI.inch
   radiusO = radiusO.inch
   txtO=@diamOutIn.to_s+"\""
   txtI=@diamInIn.to_s+"\""
end
###
Sketchup::set_status_text("Pipe Along Path... MAKING PIPE...", SB_PROMPT)
begin
	model.start_operation("Pipe Along Path", true)
rescue
	model.start_operation("Pipe Along Path")
end
@sel=ss.to_a
if @inGroup=="Yes"
	group=entities.add_group(@sel)
	@sel=group.entities.to_a
else
	group=entities.add_group()
end
###
def self.get_vertices()
### this next bit is mainly thanks to Rick Wilson's weld.rb ###
	@error=0 ### 20070512
	model=Sketchup.active_model
	ents=model.active_entities
	#@sel=model.selection
	sl=@sel.length
	verts=[]
	edges=[]
	@newVerts=[]
	startEdge=startVert=nil
#DELETE NON-EDGES, GET THE VERTICES
	@sel.each {|item| edges.push(item) if item.is_a?(Sketchup::Edge) }
	edges.each {|edge| verts.push(edge.vertices)}
	verts.flatten!
#FIND AN END VERTEX
	vertsShort=[]
	vertsLong=[]
	vertsEnds=[] ### to ensure array is only ends ###
	verts.each do |v|
		if vertsLong.include?(v)
			vertsShort.push(v)
		else
			vertsLong.push(v)
		end
	end
	vertsLong.each do |v| ### to ensure array is only ends ###
		if not vertsShort.include?(v)
			vertsEnds.push(v)
		end
	end ### 17/8/5 ### to ensure array is only ends ###
	if vertsEnds.length==0 ### i.e. it's looped ###
 		### path start or end ?
		if @theEnd==0
		  startVert=vertsLong.first
		  @startVert=@endVert=nil ###1.2
		  startEdge=startVert.edges.first
		else
		  startVert=vertsLong.last
		  @startVert=@endVert=nil ###1.2
		  startEdge=startVert.edges.first
		end
		###
		closed=true
	else
		if vertsEnds.length != 2
			Sketchup::set_status_text("Pipe Along Path... PATH ERROR ! ", SB_PROMPT)
			UI.messagebox("The selected Path either branches or is not continuous ! \nCannot make a Pipe Extrusion ! \nRe-select a single continuous path... ")
			@error=1
			return nil
		else
			### path start or end ?
			if @theEnd==0
			  startVert=vertsEnds.first
		      @startVert=startVert ###1.2
		      @endVert=vertsLong.last ###1.2
			else
			  startVert=vertsEnds.last
  		      @startVert=startVert ###1.2
			  @endVert=vertsLong.first ###1.2
			end
			###
			closed=false
			startEdge=startVert.edges.first
		end
	end
	@sel.clear
#SORT VERTICES, LIMITING TO THOSE IN THE SELECTION SET
	if startVert==startEdge.start
		@newVerts=[startVert]
		counter=0
		while @newVerts.length < verts.length
			edges.each do |edge|
				if edge.end==@newVerts.last
					@newVerts.push(edge.start)
				elsif edge.start==@newVerts.last
					@newVerts.push(edge.end)
				end
			end
			counter+=1
			if counter > verts.length
				Sketchup::set_status_text("Pipe Along Path... ERROR ! ", SB_PROMPT)
				return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
				@newVerts.reverse!
				reversed=true
			end
		end
	else
		@newVerts=[startVert]
		counter=0
		while @newVerts.length < verts.length
			edges.each do |edge|
				if edge.end==@newVerts.last
					@newVerts.push(edge.start)
				elsif edge.start==@newVerts.last
					@newVerts.push(edge.end)
				end
			end
			counter+=1
			if counter > verts.length
				Sketchup::set_status_text("Pipe Along Path... ERROR ! ", SB_PROMPT)
				return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
				@newVerts.reverse!
				reversed=true
			end
		end
	end
	@newVerts.reverse! if reversed
#CONVERT VERTICES TO POINT3Ds
	@newVerts.collect!{|x| x.position}
	@edges=edges
### now have an array of vertices in order with NO forced closed loop ...
end ### get_vertices
###
@theEnd=0
self.get_vertices()
if @error==1
  return nil
end
### NOW - do main stuff - ###
pt1 = @newVerts[0]
pt2 = @newVerts[1]
vec = pt1.vector_to(pt2)
###
### check for small arcs and == diameter bug-splatters 1.3
dflag=false
for r in arads
  dflag=true if r.inch == radiusO
end
if aflag and dflag ###1.3
  Sketchup::set_status_text("Pipe Along Path... RADIUS ? ! ", SB_PROMPT)
  UI.messagebox("The Radius of at least one Arc in the Path is < 200mm(~8\") \nand it also matches the Pipe's Outer Radius (#{radiusO}) ! \nThis will cause SketchUp to Bug-Splat and Crash !! \nHowever it CAN be done this way - \nScale up the Path and Diameters by x10, \nuse this Tool and then Scale down the Path and Pipe by x0.1...\n\nExiting...")
  return nil
end
### NOW - do main stuff - ###
entities=group.entities
if not radiusI == 0.0 ###1.1
  circleI = entities.add_circle(pt1, vec, radiusI, segs)
  baseI = entities.add_face(circleI)
end
circleO = entities.add_circle(pt1, vec, radiusO, segs)
baseO = entities.add_face(circleO)
if not radiusI == 0.0 ###1.1
   baseI.erase!
end
theFace = baseO
if pt1.x == pt2.x and pt1.y == pt2.y and pt1.z < pt2.z ### flip so not inside out ! ###
  theFace.reverse!
end
entits = model.active_entities ###1.1
theEdges= []
0.upto(@newVerts.length-2) { |i|
	if @startVert ###1.2
		theFace.reverse! if @newVerts.length < 3 and theFace.normal == [0,0,-1] and @startVert.position.z < @endVert.position.z 
	end ###1.2
	if @inGroup=="Yes"
		theEdges[i] = entities.add_line(@newVerts[i], @newVerts[i+1])  ### make vertices into edges ###1.1
	else
		theEdges[i] = entits.add_line(@newVerts[i], @newVerts[i+1])  ### make vertices into edges ###1.1
	end
}
### follow me along selected edges
theExtrusion=theFace.reverse!.followme(theEdges) ###
### add name
# group.name="OD="+txtO+" ID="+txtI # Changed to:
## Added JMcC 2013-06-12 Convert group to component 

			comp_name = UI.inputbox ["Component name"],["Scaffold pole"],"Name this component instance" 
			if comp_name # isn't nil (naming inputbox not cancelled)
				comp = group.to_component
				comp.definition.name = txtO + " OD "  + comp_name[0].to_s 
			end if

unless @clineLayer.empty?
	xlayer=layers.add(@clineLayer)
	@edges.each{|e|next unless e.valid?; e.layer=xlayer}
end
###
model.selection.clear
if @inGroup=="No"
	model.selection.add(@edges)
end
###
model.commit_operation
if @cpoints=="Yes"
	begin
		model.start_operation("Pipe's Cpoints", true)
	rescue
		model.start_operation("Pipe's Cpoints")
	end
	@newVerts.each{|v| entities.add_cpoint(v) }
	model.commit_operation ###1.3
end#if
###
end ### end def ###

### menu bits ###
unless file_loaded?(__FILE__)###
    if $submenu2
		$submenu2.add_item("Pipe Along Path"){TIG.pipe_along_path()}###
	else
		UI.menu("Plugins").add_separator###
		UI.menu("Plugins").add_item("Pipe Along Path"){TIG.pipe_along_path()}###
    end
end
file_loaded(__FILE__)###
###---------------------------------------------------------------------
end#module